import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.table import QTable

from exorad.utils.exolib import binnedPSF
from .instrument import Instrument


class Photometer(Instrument):
    """
    Photometer instrument class.
    """

    def _wavelength_table(self):
        wl_c = 0.5 * (self.description['wl_min']['value']
                      + self.description['wl_max']['value'])
        bandwidth = (self.description['wl_max']['value']
                     - self.description['wl_min']['value'])
        left_bin_edge = wl_c - 0.5 * bandwidth
        right_bin_edge = wl_c + 0.5 * bandwidth
        self.table['chName'] = [self.name]
        self.table['Wavelength'] = [wl_c.value] * wl_c.unit
        self.table['Bandwidth'] = [bandwidth.value] * bandwidth.unit
        self.table['LeftBinEdge'] = [left_bin_edge.value] * left_bin_edge.unit
        self.table['RightBinEdge'] = [right_bin_edge.value] * right_bin_edge.unit
        self.debug('wavelength table: \n{}'.format(self.table))

    def builder(self):
        self.info('building {}'.format(self.name))
        self._wavelength_table()
        # self.table['TR'], transmission_data = self._get_transmission()
        # self._add_data_to_built('transmission_data', transmission_data.to_dict())

        self.table['QE'], qe_data = self._get_qe()
        self._add_data_to_built('qe_data', qe_data.to_dict())

        if 'PSF' in self.description.keys():
            self.debug('PSF found')
            psfFilename = self.description['PSF']['value']
        else:
            psfFilename = None

        prf, pixel_rf, extent = binnedPSF(self.description['Fnum_x']['value'],
                                          self.description['Fnum_y']['value'],
                                          self.table['Wavelength'],
                                          self.description['detector']['delta_pix']['value'],
                                          filename=psfFilename)
        self._add_data_to_built('PRF', prf)
        self._add_data_to_built('pixelRF', pixel_rf)
        self._add_data_to_built('extent', extent)

        window_size_px = self.description['aperture']['radius']['value'] ** 2 * \
                         self.description['Fnum_x']['value'] * self.table['Wavelength'] * \
                         self.description['Fnum_y']['value'] * self.table['Wavelength'] / \
                         self.description['detector']['delta_pix']['value'] ** 2
        self.debug('window size : {}'.format(window_size_px))
        self._add_data_to_built('window_size_px', window_size_px)
        self.table['WindowSize'] = window_size_px

        self._add_data_to_built('table', self.table)

    def propagate_target(self, target):
        out = QTable()
        wl = target.star.sed.wl_grid
        qe, transmission, wave_window = self._get_efficiency(wl, target)
        if 'sky TR' in self.table.keys():
            out['foreground_transmission'] = self.table['sky TR']

        if self.payload['optics']['ForceChannelWlEdge']['value']:
            self.debug('force channel wl edge enabled')
            idx = np.logical_or(wl < self.description['wl_min']['value'].to(self.table['Wavelength'].unit),
                                wl > self.description['wl_max']['value'].to(self.table['Wavelength'].unit))
            transmission[idx] = wave_window[idx] = 0.0

        star_flux = np.trapz(wave_window *
                             target.star.sed.data,
                             x=target.star.sed.wl_grid
                             ).to(u.W / u.m ** 2)
        out['starFlux'] = [star_flux.value] * star_flux.unit
        self.debug('star flux : {}'.format(out['starFlux']))

        star_signal = self.payload['optics']['Atel']['value'] * np.trapz(qe *
                                                                         target.star.sed.data *
                                                                         transmission *
                                                                         target.star.sed.wl_grid.to(
                                                                             u.m) / const.c / const.h,
                                                                         x=target.star.sed.wl_grid).si * u.count
        out['starSignal'] = [star_signal.value] * star_signal.unit
        self.debug('star signal : {}'.format(out['starSignal']))

        try:
            star_signal_aperture = star_signal * self.description['aperture']['apertureCorrection']['value']
        except KeyError:
            self.debug('aperture correction not found')
            star_signal_aperture = star_signal

        out['star_signal_inAperture'] = [star_signal_aperture.value] * star_signal_aperture.unit
        self.debug('star signal in aperture : {}'.format(out['star_signal_inAperture']))

        star_signal_in_pixel = self.built_instr['PRF'].max() * star_signal
        out['star_MaxSignal_inPixel'] = [star_signal_in_pixel.value] * star_signal_in_pixel.unit
        self.debug('star signal in pixel MAX : {}'.format(out['star_signal_inAperture']))
        return out



from .task import Task


class EstimateNoiseInChannel(Task):
    """

    Parameters
    ----------
    target: Target
        target to prepare
    channel : dict
        channel to analyse
    payload : dict
        payload description

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam('target', 'target to prepare')
        self.addTaskParam('channel', 'channel to analyse')
        self.addTaskParam('payload', 'payload common section')

    def execute(self):
        from exorad.models import noise
        import numpy as np
        import astropy.units as u
        from astropy.table import QTable
        channel = self.get_task_param('channel')
        target = self.get_task_param('target')
        payload = self.get_task_param('payload')

        name = channel['value']
        self.info('estimating noise in {}'.format(name))

        out = QTable()
        out = noise.frame_time(target, channel, out)
        read_gain, shot_gain = noise.multiaccum(channel, out['frameTime'][0])
        self.debug('read gain : {} , shot gain : {}'.format(read_gain, shot_gain))

        out = noise.photon_noise(target.table, channel, shot_gain, out)

        out['darkcurrent_noise'] = np.sqrt(shot_gain * target.table['WindowSize'][target.table['chName'] == name] *
                                           channel['detector']['dark_current']['value'] * u.ct / u.hr).to(
            u.ct / u.s) * u.hr ** 0.5
        self.debug('dark current noise : {}'.format(out['darkcurrent_noise']))
        out['read_noise'] = np.sqrt(
            read_gain * channel['detector']['read_noise']['value'] ** 2 *
            target.table['WindowSize'][target.table['chName'] == name] / out['frameTime'] * 1 / u.hr).to(
            u.ct / u.s) * u.hr ** 0.5
        self.debug('read noise : {}'.format(out['read_noise']))

        if 'NoiseX' in channel:
            NoiseX = channel['NoiseX']['value']
        else:
            NoiseX = 0.0
        self.debug('NoiseX: {}'.format(NoiseX))

        photon_noise_variance = noise.photon_noise_variance(target.table, out)
        signal = target.table['star_signal_inAperture'][target.table['chName'] == name].copy()
        signal[signal == 0.0] = np.nan
        out['total_noise'] = np.sqrt(
                                out['darkcurrent_noise'] ** 2 +
                                (1.0 + NoiseX) *
                                (photon_noise_variance) +
                                out['read_noise'] ** 2
                            ) / signal

        wl = target.table['Wavelength'][target.table['chName'] == name]
        if payload:
            try:
                out = noise.add_custom_noise(payload['common']['customNoise'], wl, out)
            except KeyError:
                self.debug('no custom noise found in common section')
        try:
            out = noise.add_custom_noise(channel['customNoise'], wl, out)
        except KeyError:
            self.debug('no custom noise found in channel section')

        self.debug('total noise : {}'.format(out['total_noise']))

        self.set_output(out)


class EstimateNoise(Task):
    """

    Parameters
    ----------
    target: Target
        target to prepare
    channels : dict
        channel list to analyse

    Returns
    -------
    Target:
        same target with table attribute updated
    """

    def __init__(self):
        self.addTaskParam('target', 'target to prepare')
        self.addTaskParam('channels', 'channel list to analyse')

    def execute(self):
        from exorad.utils.util import vstack_tables
        from exorad.tasks import UpdateTargetTable
        from exorad.tasks import EstimateMaxSignal

        estimateMaxSignal = EstimateMaxSignal()
        estimateNoiseInChannel = EstimateNoiseInChannel()
        updateTargetTable = UpdateTargetTable()

        self.info('computing noise')
        target = self.get_task_param('target')
        channels = self.get_task_param('channels')

        target = estimateMaxSignal(target=target)

        table_list = []
        for ch in self.get_task_param('channels'):
            self.debug('computing noise in {}'.format(ch))
            table_list.append(
                estimateNoiseInChannel(target=target, channel=channels[ch].description, payload=channels[ch].payload))
        table = vstack_tables(table_list)
        new_target = updateTargetTable(target=target, table=table)
        self.set_output(new_target)



from .foregroundHandler import EstimateZodi, EstimateForeground, EstimateForegrounds
from .instrumentHandler import BuildChannels, BuildInstrument, LoadPayload, PreparePayload, MergeChannelsOutput, \
    GetChannelList
from .loadOptions import LoadOptions
from .loadSource import LoadSource
from .noiseHandler import EstimateNoise, EstimateNoiseInChannel
from .propagateLight import PropagateTargetLight, PropagateForegroundLight
from .targetHandler import LoadTargetList, PrepareTarget, UpdateTargetTable, EstimateMaxSignal, \
    ObserveTarget, ObserveTargetlist



import logging
import os

import astropy.units as u
import mpmath
import numpy as np
from astropy.io import fits
from scipy import signal
from scipy.integrate import cumtrapz
from scipy.special import j1
from scipy.stats import binned_statistic

logger = logging.getLogger('exorad.exolib')


def rebin(x, xp, fp):
    ''' Resample a function fp(xp) over the new grid x, rebinning if necessary,
      otherwise interpolates
      Parameters
      ----------
      x	: 	array like
      New coordinates
      fp 	:	array like
      y-coordinates to be resampled
      xp 	:	array like
      x-coordinates at which fp are sampled

      Returns
      -------
      out	: 	array like
      new samples

    '''

    if (x.unit != xp.unit):
        logger.fatal('Units mismatch in rebin {:s}, {:s}'.format(x.unit, xp.unit))
        raise ValueError

    idx = np.where(np.logical_and(xp > 0.9 * x.min(), xp < 1.1 * x.max()))[0]
    xp = xp[idx]
    fp = fp[idx]

    if np.diff(xp).min() < np.diff(x).min():
        # Binning!
        bin_x = 0.5 * (x[1:] + x[:-1])
        x0 = x[0] - (bin_x[0] - x[0]) / 2.0
        x1 = x[-1] + (x[-1] - bin_x[-1]) / 2.0
        bin_x = np.insert(bin_x, [0], x0)
        bin_x = np.append(bin_x, x1)
        new_f = binned_statistic(xp, fp, bins=bin_x, statistic='mean')[0] * fp.unit

        idx = np.where(np.logical_and(xp >= x.min(), xp <= x.max()))[0]

    #    print np.trapz(fp[idx], x=xp[idx])
    #    print np.trapz(new_f, x=x)
    #    asd
    else:
        # Interpolate !
        new_f = np.interp(x, xp, fp, left=0.0, right=0.0)
        #         from scipy.interpolate import interp1d
        #         interpolator = interp1d(xp, fp, kind='cubic', fill_value=0.0, assume_sorted=False, bounds_error=False)
        #         new_f = interpolator(x)

        if not isinstance(new_f, u.Quantity):
            new_f *= fp.unit

    return x, new_f


def rebin_(x, xp, fp):
    ''' Resample a function fp(xp) over the new grid x, rebinning if necessary,
      otherwise interpolates
      Parameters
      ----------
      x	: 	array like
      New coordinates
      fp 	:	array like
      y-coordinates to be resampled
      xp 	:	array like
      x-coordinates at which fp are sampled

      Returns
      -------
      out	: 	array like
      new samples

    '''

    if (x.unit != xp.unit):
        logger.fatal('Units mismatch in rebin {:s}, {:s}'.format(x.unit, xp.unit))
        raise ValueError

    idx = np.where(np.logical_and(xp > 0.9 * x.min(), xp < 1.1 * x.max()))[0]
    xp = xp[idx]
    fp = fp[idx]

    if np.diff(xp).min() < np.diff(x).min():
        # Binning!
        c = cumtrapz(fp.value, x=xp.value) * fp.unit * xp.unit
        xpc = xp[1:]

        delta = np.gradient(x)
        new_c_1 = np.interp(x - 0.5 * delta, xpc, c,
                            left=0.0, right=0.0) * c.unit
        new_c_2 = np.interp(x + 0.5 * delta, xpc, c,
                            left=0.0, right=0.0) * c.unit
        new_f = (new_c_2 - new_c_1) / delta
    else:
        # Interpolate !
        new_f = np.interp(x, xp, fp, left=0.0, right=0.0) * fp.unit

    return x, new_f


def planck(wl, T):
    """ Planck function.

      Parameters
      __________
        wl : 			array
                  wavelength [micron]
        T : 			scalar
                  Temperature [K]
                  Spot temperature [K]
      Returns
      -------
        spectrum:			array
                  The Planck spectrum  [W m^-2 sr^-2 micron^-1]
    """

    a = np.float64(1.191042768e8) * u.um ** 5 * u.W / u.m ** 2 / u.sr / u.um
    b = np.float64(14387.7516) * 1 * u.um * u.K
    try:
        x = b / (wl * T)
        bb = a / wl ** 5 / (np.exp(x) - 1.0)
    except ArithmeticError:
        bb = np.zeros_like(wl)
    return bb


def binnedPSF(F_x, F_y, wl, delta_pix, filename=None, PSFtype='AIRY'):
    if filename:
        with fits.open(os.path.expanduser(filename)) as hdu:
            ima = hdu[0].data
            hdr = hdu[0].header
            # define a kernel representing the detector pixel response
            # and use fractional pixel
            k_x = delta_pix / (F_x * wl * hdr['CDELT2'])
            k_y = delta_pix / (F_y * wl * hdr['CDELT1'])
            xmin = -(hdr['CRPIX2'] - 1) * F_x * wl * hdr['CDELT2']
            xmax = (hdr['NAXIS2'] - (hdr['CRPIX2'] - 1)) * F_x * wl * hdr['CDELT2']
            ymin = -(hdr['CRPIX1'] - 1) * F_y * wl * hdr['CDELT1']
            ymax = (hdr['NAXIS1'] - (hdr['CRPIX1'] - 1)) * F_y * wl * hdr['CDELT1']
            extent = (xmin, xmax, ymin, ymax)
    elif PSFtype == 'AIRY':
        x = np.linspace(-3.0, 3.0, 256)
        xx, yy = np.meshgrid(x, x)
        r = np.pi * np.sqrt(xx ** 2 + yy ** 2) + 1.0e-10

        ima = (2.0 * j1(r) / r) ** 2
        ima *= 0.25 * np.pi * (x[1] - x[0]) ** 2
        # print ima.sum()
        k_x = delta_pix / (F_x * wl * (x[1] - x[0]))
        k_y = delta_pix / (F_y * wl * (x[1] - x[0]))
        # print k_y, k_x
        extent = (-(ima.shape[1] // 2) * F_x * wl * (x[1] - x[0]),
                  (ima.shape[1] // 2) * F_x * wl * (x[1] - x[0]),
                  -(ima.shape[0] // 2) * F_y * wl * (x[1] - x[0]),
                  (ima.shape[0] // 2) * F_y * wl * (x[1] - x[0]))

    fk_x, ik_x = np.modf(k_x)
    fk_y, ik_y = np.modf(k_y)

    kernel = np.ones((int(ik_y) + 2, int(ik_x) + 2)) * fk_x.unit
    kernel[:, 0] *= 0.5 * fk_x
    kernel[:, -1] *= 0.5 * fk_x
    kernel[0, :] *= 0.5 * fk_y
    kernel[-1, :] *= 0.5 * fk_y

    imac = signal.convolve2d(ima, kernel, mode='same')

    return imac, kernel, extent


def OmegaPix(Fnum_x, Fnum_y=None):
    ''' 
    Calculate the solid angle subtended by an elliptical aperture on-axis.
    Algorithm from "John T. Conway. Nuclear Instruments and Methods in 
    Physics Research Section A: Accelerators, Spectrometers, Detectors and
    Associated Equipment, 614(1):17 ? 27, 2010. 
    https://doi.org/10.1016/j.nima.2009.11.075
    Equation n. 56
    
    Parameters
    ----------
    Fnum_x : scalar
             Input F-number along dispersion direction
    Fnum_y : scalar
             Optional, input F-number along cross-dispersion direction
             
    Returns
    -------
     Omega : scalar
             The solid angle subtanded by an elliptical aperture in units u.sr
             
    '''

    if not Fnum_y: Fnum_y = Fnum_x

    if Fnum_x > Fnum_y:
        a = 1.0 / (2 * Fnum_y)
        b = 1.0 / (2 * Fnum_x)
    else:
        a = 1.0 / (2 * Fnum_x)
        b = 1.0 / (2 * Fnum_y)

    h = 1.0

    A = 4.0 * h * b / (a * np.sqrt(h ** 2 + a ** 2))
    k = np.sqrt((a ** 2 - b ** 2) / (h ** 2 + a ** 2))
    alpha = np.sqrt(1 - (b / a) ** 2)

    Omega = 2.0 * np.pi - A * mpmath.ellippi(alpha ** 2, k ** 2)

    return Omega * u.sr


if __name__ == "__main__":
    Omega = OmegaPix(5.0, 5.0)

    sigma = np.arctan(1.0 / 10)
    OmegaCheck = 2.0 * np.pi * (1 - np.cos(sigma))
    print((Omega, OmegaCheck))



import logging

import astropy.constants as const
import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d

from exorad.models.signal import Signal
from exorad.utils.exolib import OmegaPix

logger = logging.getLogger('exorad.diffuse light')


def prepare(ch_table, ch_built_instr, description):
    logger.info('computing signal')
    wl_table = ch_table['Wavelength']
    logger.debug('wl table : {}'.format(wl_table))

    fnum_x = description['Fnum_x']['value'].value
    if 'Fnum_y' in description.keys():
        fnum_y = description['Fnum_y']['value'].value
    else:
        fnum_y=None
    omega_pix = OmegaPix(fnum_x,fnum_y)
    logger.debug('omega pix : {}'.format(omega_pix))
    A = (description['detector']['delta_pix']['value']
         * description['detector']['delta_pix']['value']).to(u.m ** 2)
    qe = Signal(ch_built_instr['qe_data']['wl_grid']['value'] * u.Unit(ch_built_instr['qe_data']['wl_grid']['unit']),
                ch_built_instr['qe_data']['data']['value'])
    logger.debug('wl qe : {}'.format(qe.wl_grid))

    transmission = Signal(ch_built_instr['transmission_data']['wl_grid']['value'] * \
                          u.Unit(ch_built_instr['transmission_data']['wl_grid']['unit']),
                          ch_built_instr['transmission_data']['data']['value'])

    total_signal = np.zeros(wl_table.size) * u.count / u.s
    total_max_signal = np.zeros(wl_table.size) * u.count / u.s
    return total_max_signal, total_signal, wl_table, A, qe, omega_pix, transmission




def convolve_with_slit(ch_description, ch_built_instr, A, ch_table, omega_pix, qe, radiance):
    slit_width = ch_built_instr['slit_width']
    wl_pix = ch_built_instr['wl_pix_center']
    dwl_pic = ch_built_instr['pixel_bandwidth']
    radiance.spectral_rebin(wl_pix)
    logger.debug('radiance : {}'.format(radiance.data))

    qe_func = interp1d(qe.wl_grid,
                       qe.data,
                       assume_sorted=False,
                       fill_value=0.0,
                       bounds_error=False)
    aomega = omega_pix * A * qe_func(wl_pix) * (wl_pix / const.c / const.h).to(1. / u.W / u.s) * u.count
    logger.debug('AOmega : {}'.format(aomega))
    radiance.data *= aomega
    logger.debug('sed : {}'.format(radiance.data))
    logger.debug('convolving with slit')
    slit_kernel = np.ones(int(slit_width / ch_description['detector']['delta_pix']['value'].to(u.um)))
    signal_tmp = (np.convolve(radiance.data * dwl_pic, slit_kernel, 'same')).to(u.count / u.s)
    logger.debug('signal_tmp: {}'.format(signal_tmp))
    idx = [np.logical_and(wl_pix > wlow, wl_pix <= whigh)
           for wlow, whigh in zip(ch_table['LeftBinEdge'], ch_table['RightBinEdge'])]
    signal = [signal_tmp[idx[k]].sum() for k in range(len(idx))]
    signal = [sig * w for sig, w in zip(signal, np.array(ch_built_instr['window_spatial_width']))]
    logger.debug('signal: {}'.format(signal))
    try:
        max_signal_per_pix = [signal_tmp[idx[k]].max() for k in range(len(idx))]
    except ValueError:
        logger.error('no max value in the array')
        raise
    return max_signal_per_pix, signal


def integrate_light(radiance, wl_qe, ch_built_instr):
    logger.debug('sed : {}'.format(radiance.data))
    signal_tmp = np.trapz(radiance.data, x=wl_qe).to(u.count / u.s)
    # signal_tmp = (np.trapz(radiance.data[~np.isnan(radiance.data)], x=wl_qe[~np.isnan(radiance.data)])).to(
    #     u.count / u.s)
    signal = signal_tmp * np.asarray(ch_built_instr['window_size_px'])
    max_signal_per_pix = signal_tmp * np.ones_like(ch_built_instr['window_size_px'])
    return max_signal_per_pix, signal



import logging
import os
import pathlib
import shutil

import matplotlib.pyplot as plt

import exorad.__version__ as version
import exorad.tasks as tasks
from exorad.log import setLogLevel, addLogFile
from exorad.output.hdf5 import HDF5Output
from exorad.utils.plotter import Plotter

logger = logging.getLogger('exorad')

preparePayload = tasks.PreparePayload()
mergeChannelsOutput = tasks.MergeChannelsOutput()
loadTargetList = tasks.LoadTargetList()
loadSource = tasks.LoadSource()
observeTarget = tasks.ObserveTarget()
observeTargetList = tasks.ObserveTargetlist()


def efficiency_plot(channels, output_dir=None):
    table = mergeChannelsOutput(channels=channels)
    plotter = Plotter(channels=channels, input_table=table)
    plotter.plot_efficiency()
    if output_dir is None:
        plt.show()
    else:
        plotter.save_fig(os.path.join(output_dir, 'efficiency.png'), efficiency=True)
        logger.info('efficiency plot saved in output directory')
    plt.close()


def standard_pipeline(options, target_list, output=None, plot=False, full_contrib=False,
                      n_thread=1, debug=False, log=False):
    from exorad.utils.ascii_art import ascii_art
    logger.info(ascii_art)
    logger.info('code version {}'.format(version))

    if debug: setLogLevel(logging.DEBUG)
    if isinstance(log, str): addLogFile(fname=log)
    elif log: addLogFile()

    if output is not None:
        out_dir = pathlib.Path(output).parent.absolute()
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
            logger.info('output directory created')
        logger.info('output directory set as {}'.format(out_dir))
        try:
            shutil.copy(options, out_dir)
        except shutil.SameFileError:
            pass
        try:
            shutil.copy(target_list, out_dir)
        except shutil.SameFileError:
            pass
    else:
        out_dir = None

    # step 1 load payload
    payload, channels, (wl_min, wl_max) = preparePayload(payload_file=options,
                                                         output=output)
    if full_contrib:
        from astropy.table import hstack
        for ch in channels:
            channels[ch].table = hstack([channels[ch].table, channels[ch].built_instr['optical_path']['signal_table']])
            channels[ch].table.remove_column('instrument_signal')

    # step 1b plot payload efficiency
    if plot:
        efficiency_plot(channels=channels, output_dir=out_dir)

    # step 2 load targetlist
    targets = loadTargetList(target_list=target_list)

    # step 3 observe targetlist
    targets = observeTargetList(targets=targets.target, payload=payload, channels=channels, wl_range=(wl_min, wl_max),
                                plot=plot, out_dir=out_dir, n_thread=n_thread, debug=debug)
    # step 4 save to output
    if output is not None:
        for target in targets:
            with HDF5Output(output, append=True) as out:
                targets[target].write(out)


def main():
    import argparse

    parser = argparse.ArgumentParser(description='ExoRad {}'.format(version))
    parser.add_argument("-p", "--payload", dest='opt', type=str,
                        required=True, help="Input payload description file to pass")
    parser.add_argument("-t", "--targetList", dest='targetList', type=str,
                        required=True, help="Input target list file to pass")
    parser.add_argument("-o", "--output", dest='output', type=str,
                        required=False, default=None, help="Output file")
    parser.add_argument("-c", "--full_contrib", dest='contrib', default=False,
                        required=False, help="produce full contribution output", action='store_true')
    parser.add_argument("-n", "--nThreads", dest='numberOfThreads', default=1, type=int,
                        required=False, help="number of threads for parallel processing")
    parser.add_argument("-d", "--debug", dest='debug', default=False,
                        required=False, help="log output on screen", action='store_true')
    parser.add_argument("-l", "--log", dest='log', default=False,
                        required=False, help="log output on file", action='store_true')
    parser.add_argument("-P", "--plot", dest='plot', default=False,
                        required=False, help="save target plots", action='store_true')

    args = parser.parse_args()

    standard_pipeline(options=args.opt, target_list=args.targetList,
                      output=args.output, plot=args.plot, full_contrib=args.contrib,
                      debug=args.debug, n_thread=args.numberOfThreads, log=args.log)



"""
These parts of the code derives from TauREx3:
    @ARTICLE{2019arXiv191207759A,
           author = {{Al-Refaie}, Ahmed F. and {Changeat}, Quentin and {Waldmann}, Ingo P. and {Tinetti}, Giovanna},
            title = "{TauREx III: A fast, dynamic and extendable framework for retrievals}",
          journal = {arXiv e-prints},
         keywords = {Astrophysics - Instrumentation and Methods for Astrophysics, Astrophysics - Earth and Planetary Astrophysics},
             year = 2019,
            month = dec,
              eid = {arXiv:1912.07759},
            pages = {arXiv:1912.07759},
    archivePrefix = {arXiv},
           eprint = {1912.07759},
     primaryClass = {astro-ph.IM},
           adsurl = {https://ui.adsabs.harvard.edu/abs/2019arXiv191207759A},
          adsnote = {Provided by the SAO/NASA Astrophysics Data System}
    }

BSD 3-Clause License

Copyright (c) 2019, Ahmed F. Al-Refaie, Quentin Changeat, Ingo Waldmann, Giovanna Tinetti
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the names of the copyright holders nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from abc import ABC

import numpy as np

from exorad.log import Logger


class Output(Logger):

    def __init__(self):
        super().__init__()

    def open(self):
        raise NotImplementedError

    def create_group(self, group_name):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def store_dictionary(self, dictionary, group_name=None):
        from exorad.output.hdf5.util import recursively_save_dict_contents_to_output

        out = self
        if group_name is not None:
            out = self.create_group(group_name)

        recursively_save_dict_contents_to_output(out, dictionary)


class OutputGroup(Output, ABC):

    def __init__(self, name):
        super().__init__(name)
        self._name = name

    def write_array(self, array_name, array, metadata=None):
        raise NotImplementedError

    def write_list(self, list_name, list_array, metadata=None):
        arr = np.array(list_array)
        self.write_array(list_name, arr)

    def write_scalar(self, scalar_name, scalar, metadata=None):
        raise NotImplementedError

    def write_string(self, string_name, string, metadata=None):
        raise NotImplementedError

    def write_string_array(self, string_name, string_array, metadata=None):
        raise NotImplementedError



import astropy.units as u
import numpy as np
from astropy.table import Table, QTable, meta, serialize

from exorad.models.signal import Signal, Sed

signal = {'W / (m2 um)': Sed}


def recursively_read_dict_contents(input_dict):
    """
    Will recursive read a dictionary, initializing quantities and table from a dictionary read from an hdf5 file.

    Parameters
    ----------
    input_dict : dict
        dictionary read from hdf5

    Returns
    --------
    dict
        Dictionary we want to use

    """
    new_keys = [k for k in input_dict.keys()]
    # if all(elem in new_keys for elem in ['wl_grid', 'data', 'time_grid']):
    #     wl = input['wl_grid']['value'] * u.Unit(input['wl_grid']['unit'])
    #     data = input['data']['value'] * u.Unit(input['data']['unit'])
    #     time = input['time_grid']['value'] * u.Unit(input['time_grid']['unit'])
    #     input = signal[str(input['data']['unit'])](wl, data, time)
    if all(elem in new_keys for elem in ['value', 'unit']):
        input_dict['value'] = input_dict['value'] * u.Unit(input_dict['unit'])

    if any('.__table_column_meta__' in elem for elem in new_keys):
        table_keys = [elem for elem in new_keys if '.__table_column_meta__' in elem]
        table_keys = (elem.split('.')[0] for elem in table_keys)
        for k in table_keys:
            table = load_table(input_dict, k)
            input_dict[k] = table
    for key in new_keys:
        if isinstance(input_dict[key], dict):
            input_dict[key] = recursively_read_dict_contents(input_dict[key])
    return input_dict


def load_table(input_table, k):
    table = Table(np.array(input_table[k]))
    header = meta.get_header_from_yaml(h.decode('utf-8') for h in input_table['{}.__table_column_meta__'.format(k)])
    header_cols = dict((x['name'], x) for x in header['datatype'])
    for col in table.columns.values():
        for attr in ('description', 'format', 'unit', 'meta'):
            if attr in header_cols[col.name]:
                setattr(col, attr, header_cols[col.name][attr])
    table = serialize._construct_mixins_from_columns(table)
    try:
        header['meta'].pop('__serialized_columns__')
        table.meta = header['meta']
    except KeyError:
        pass
    return table


def recursively_save_dict_contents_to_output(output, dic):
    """
    Will recursive write a dictionary into output.

    Parameters
    ----------
    output :
        Group (or root) in output file to write to

    dic : :obj:`dict`
        Dictionary we want to write

    """

    for key, item in dic.items():
        try:
            store_thing(output, key, item)
        except TypeError:
            raise ValueError('Cannot write %s type' % type(item))
    return


def store_thing(output, key, item):
    if isinstance(item, u.Quantity):
        output.write_quantity(key, item)
    elif isinstance(item, (float, int, np.int64, np.float64,)):
        output.write_scalar(key, item)
    elif isinstance(item, np.ndarray):
        if True in [isinstance(x, str) for x in item]:
            output.write_string_array(key, item)
        else:
            output.write_array(key, item)
    elif isinstance(item, (str,)):
        output.write_string(key, item)
    elif isinstance(item, (Table, QTable)):
        output.write_table(key, item)
    elif isinstance(item, Signal):
        group = output.create_group(key)
        recursively_save_dict_contents_to_output(group, item.to_dict())
    elif isinstance(item, (list, tuple,)):
        if isinstance(item, tuple):
            item = list(item)
        if True in [isinstance(x, str) for x in item]:
            output.write_string_array(key, item)
        else:
            try:
                output.write_array(key, np.array(item))

            except TypeError:
                for idx, val in enumerate(item):
                    new_key = '{}{}'.format(key, idx)
                    store_thing(output, new_key, val)

    elif isinstance(item, dict):
        group = output.create_group(key)
        recursively_save_dict_contents_to_output(group, item)
    else:
        raise TypeError
    return



import codecs
import os

from setuptools import setup, find_packages

packages = find_packages(exclude=('tests', 'docs'))

provides = ['exorad', ]
console_scripts = [
    'exorad=exorad.exorad:main',
    'exorad-plot=exorad.utils.plotter:main [Plot]']

install_requires = ['astropy',
                    'matplotlib',
                    'seaborn',
                    'mpmath',
                    'numpy',
                    'pyyaml',
                    'scipy',
                    'h5py',
                    'hdfdict',
                    'xlrd==1.2.0',
                    'xlwt'
                    ]
entry_points = {'console_scripts': console_scripts, }

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Unix',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: Libraries',
]


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_info(rel_path, info):
    for line in read(rel_path).splitlines():
        if line.startswith('__%s__' % info):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find %s string." % info)


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='exorad',
      provides=provides,
      version=get_info("exorad/__version__.py", 'version'),
      description=get_info("exorad/__init__.py", 'summary'),
      url=get_info("exorad/__init__.py", 'url'),
      author=get_info("exorad/__init__.py", 'author'),
      author_email=get_info("exorad/__init__.py", 'email'),
      license=get_info("exorad/__init__.py", 'license'),
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=packages,
      classifiers=classifiers,
      install_requires=install_requires,
      include_package_data=True,
      entry_points=entry_points,
      python_requires='>=3.8',
      zip_safe=False)



from .targetHandler import UpdateTargetTable
from .task import Task


class PropagateTargetLight(Task):
    """
     Given the target and a channel dict, it propagates the target light over the channels

     Parameters
     ----------
     channels: dict
         main payload. Default is None
     target: Target

     Returns
     -------
     QTable:
         return the output table of propagated signal

     Examples
     --------

     """

    def __init__(self):
        self.addTaskParam('channels', 'payload channels dict')
        self.addTaskParam('target', 'Target class to observe')

    def execute(self):
        self.info('propagating target light')
        from exorad.utils.util import vstack_tables
        target = self.get_task_param('target')
        channels = self.get_task_param('channels')
        self.debug('detectors found : {}'.format(channels.keys()))
        table_list = []
        for ch in self.get_task_param('channels'):
            self.debug('propagating target in {}'.format(ch))
            table_list.append(channels[ch].propagate_target(target))
        table = vstack_tables(table_list)
        updateTargetTable = UpdateTargetTable()
        target = updateTargetTable(target=target, table=table)
        self.set_output(target)


class PropagateForegroundLight(Task):
    """
     Given the target and a channel dict, it propagates the target foreground light over the channels

     Parameters
     ----------
     channels: dict
         main payload. Default is None
     target: Target

     Returns
     -------
     QTable:
         return the output table of propagated signal

     Examples
     --------

     """

    def __init__(self):
        self.addTaskParam('channels', 'payload channels dict')
        self.addTaskParam('target', 'Target class to observe')

    def execute(self):
        self.info('propagating target foreground light')
        from exorad.utils.util import vstack_tables
        target = self.get_task_param('target')
        channels = self.get_task_param('channels')
        self.debug('detectors found : {}'.format(channels.keys()))
        table_list = []
        for ch in self.get_task_param('channels'):
            self.debug('propagating target background in {}'.format(ch))
            table_list.append(channels[ch].propagate_diffuse_foreground(target))
        table = vstack_tables(table_list)
        updateTargetTable = UpdateTargetTable()
        target = updateTargetTable(target=target, table=table)
        self.set_output(target)
